# include "std.h"
# if LICENCE
   This file is part of cwrite.
   Copyright 2007, Desislav Georgiev.
 
   cwrite is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   cwrite is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with cwrite.  If not, see <http://www.gnu.org/licenses/>.
# endif
# if HEADER
   # include <stdio.h>
   # include <stdlib.h>

   # include "block.h"
   # include "command_line.h"
   # include "cursor.h"
   # include "curses.h"
   # include "cwrite.h"
   # include "datachain.h"
   # include "dataline.h"
   # include "datasystem.h"
   # include "file.h"
   # include "message.h"
   # include "preferences.h"
   # include "screen.h"
   # include "scroll.h"
   # include "status_line.h"
   # include "string.h"
# endif
# if GLOBAL
   int * _dataline_offset;
   int dataline_start_size;
   int dataline_extend_size;
# endif
# if PROTO
   void dataline_shift_right();
# endif

void dataline_fill(int c)
   { if ( * _status_register & INS_MODE )
        { if ( (* __datachain)->size == _datasystem->dataline_max_size )
             { datasystem_realloc();
             }
        }
     //else
        { if ( * _dataline_offset == _datasystem->dataline_max_size )
             { datasystem_realloc();
             }
        }
     char * _dataline = (* __datachain)->_line;
     
     if ( * _status_register & INS_MODE )
        { if ( * _dataline_offset < (* __datachain)->size )
             { dataline_shift_right();
             }
        }
     * (_dataline + * _dataline_offset) = c;
     (* _dataline_offset)++;
     if ( * _dataline_offset >= (* __datachain)->size )
        { (* __datachain)->size = * _dataline_offset;
        }
     * _status_register |= MODIFY_STATE;
     return;
   }
void dataline_fill(int c, int n)
   { if ( ! (* _status_register & TYPE_MODE) )
        { return;
        }
     for ( int i = 0; i < n; i++ )
        { dataline_fill(c);
          if ( * _cursor_x < screen_right )
             { (* _cursor_x)++;
             }
        }
     screen_redraw();

     return;
   }

int dataline_begin(Datachain * _datachain)
   { char * _dataline_point = _datachain->_line;
     if ( ! _datachain->size )
        { return 0;
        }
     for (int i = 0; i < _datachain->size; i++)
        { if ( * _dataline_point != ' ' )
             { break;
             }
          _dataline_point++;
        }
     if ( _dataline_point == _datachain->_line + _datachain->size )
        { return 0;
        }
     if ( * _dataline_point == '{' )
        { if ( _dataline_point < _datachain->_line + _datachain->size - 2 )
             { _dataline_point += 2;
             }
        }
     if ( * _dataline_point == '}' )
        { if ( _dataline_point >= _datachain->_line + _preferences->tab_value )
             { _dataline_point -= _preferences->tab_value;
             }
          else
             { _dataline_point = _datachain->_line;
             }
        }
     return (_dataline_point - _datachain->_line);
   }

void dataline_insert()
   { Datachain * _datachain_next = (* __datachain)->_next;

     if ( ! (* _status_register & TYPE_MODE) )
        { return;
        }
     if ( _datachain_next && _datachain_next->invisible )
        { if ( seek_visible_line_down() )
             { * __datachain = (* __datachain)->_prev;
             }
          else
             { if ( * _status_register & DETACH_STATE )
                  { * __datachain = * __datachain_detach_end;
                  }
               else
                  { * __datachain = * __datachain_last;
                  }
             }
        }
     * __datachain = (Datachain *) _datachain_insert(_datasystem, * __datachain);
     * _dataline_offset = 0;
     if ( * _cursor_y < screen_bottom )
        { (* _cursor_y)++;
        }
     * _cursor_x = screen_left;
     screen_redraw();
     * _status_register |= MODIFY_STATE;

     return;
   }
void dataline_insert_smart()
   { Datachain * _datachain_next = (* __datachain)->_next;
     Datachain * _datachain_prev;

     if ( ! (* _status_register & TYPE_MODE) )
        { return;
        }
     if ( _datachain_next && _datachain_next->invisible )
        { if ( seek_visible_line_down() )
             { _datachain_prev = (* __datachain)->_prev;
             }
          else
             { if ( * _status_register & DETACH_STATE )
                  { _datachain_prev = * __datachain_detach_end;
                  }
               else
                  { _datachain_prev = * __datachain_last;
                  }
             }
        }
     else
        { _datachain_prev = * __datachain;
        }
     * __datachain = (Datachain *) _datachain_insert(_datasystem, _datachain_prev);
     int start_position = dataline_begin(_datachain_prev);
     * _dataline_offset = start_position;
     if ( * _cursor_y < screen_bottom )
        { (* _cursor_y)++;
        }
     if ( start_position + screen_left < screen_cols )
        { * _cursor_x = start_position + screen_left;
        }
     else
        { * _cursor_x = 1;
        }
     screen_redraw();
     * _status_register |= MODIFY_STATE;

     return;
   }

void dataline_shift_left(Datachain * _datachain, int n)
   { char * _dataline = _datachain->_line;
     char * _dataline_point = _dataline + * _dataline_offset;

        // -- shift n symbols left --
     for ( int i = 0; i < _datachain->size - * _dataline_offset - n; i++ )
        { * (_dataline_point + i) = * (_dataline_point + n + i);
        }

        // -- set n spaces --
     _datachain->size -= n;
     _dataline_point = _dataline + _datachain->size;
     for ( int i = 0; i < n; i++ )
        { * _dataline_point++ = ' ';
        }
     * _status_register |= MODIFY_STATE;
     return;
   }
void dataline_shift_right()
   { Datachain * _datachain = * __datachain;
     char * _dataline = _datachain->_line;
     char * _dataline_point = _dataline + _datachain->size;

     for ( int i = 0; i < _datachain->size - * _dataline_offset; i++ )
        { * _dataline_point = * (_dataline_point - 1);
          _dataline_point--;
        }
     _datachain->size++;
     * _status_register |= MODIFY_STATE;
     return;
   }

void selection_shift(Datachain * _datachain)
   { char * _dataline = _datachain->_line;
     char * _dataline_point = _dataline;
     if ( _datachain->selection_end > _datachain->size )
        { _datachain->selection_end = _datachain->size;
        }
     int selection_offset = _datachain->selection_end - _datachain->selection_start;

     for ( int i = 0; i <= _datachain->size - _datachain->selection_end; i++ )
        { * (_dataline_point + _datachain->selection_start + i) = * (_dataline_point + _datachain->selection_end + i);
        }
     _dataline_point = _dataline + _datachain->size - selection_offset;
     for (int i = 0; i < selection_offset; i++)
        { * _dataline_point++ = ' ';
        }
     _datachain->size -= selection_offset;
     * _status_register |= MODIFY_STATE;
     return;
   }
void selection_shift_left(Datachain * _datachain)
   { char * _dataline = _datachain->_line;
     char * _dataline_point = _dataline + _datachain->selection_start - 1;

     for ( int i = 0; i < _datachain->selection_end - _datachain->selection_start; i++ )
        { * _dataline_point = * (_dataline_point + 1);
          _dataline_point++;
        }
     * _dataline_point = ' ';
     _datachain->selection_start--;
     _datachain->selection_end--;
     _datachain->size--;
     * _status_register |= MODIFY_STATE;
     return;
   }
void selection_shift_right(Datachain * _datachain)
   { char * _dataline = _datachain->_line;
     char * _dataline_point = _dataline + _datachain->selection_end;

     for ( int i = 0; i < _datachain->selection_end - _datachain->selection_start; i++ )
        { * _dataline_point = * (_dataline_point - 1);
          _dataline_point--;
        }
     * _dataline_point = ' ';
     _datachain->selection_start++;
     _datachain->selection_end++;
     _datachain->size++;
     * _status_register |= MODIFY_STATE;
     return;
   }

void dataline_delete_to_beg()
   { if ( ! * _dataline_offset )
        { return;
        }
     int n = * _dataline_offset;
     cursor_home();
     dataline_shift_left(* __datachain, n);
     * _status_register |= MODIFY_STATE;
     screen_line_redraw(* __datachain, * _cursor_y);
     return;
   }
void dataline_delete_to_eol()
   { if ( ! (* _status_register & TYPE_MODE) )
        { return;
        }
     Datachain * _datachain = * __datachain;
     if ( ! _datachain->size )
        { return;
        }
     if ( * _dataline_offset >= _datachain->size )
        { return;
        }
     _memset(_datachain->_line + * _dataline_offset, ' ', _datachain->size - * _dataline_offset);

     _datachain->size = * _dataline_offset;
     * _status_register |= MODIFY_STATE;
     screen_redraw();
     return;      
   }

void block_remove(Datachain * _datachain_begin, Datachain * _datachain_end, bool insert_line_rq)
   { if ( ! (* _status_register & TYPE_MODE) )
        { return;
        }

     Datachain * _datachain_prev = _datachain_begin->_prev;
     Datachain * _datachain_next;

     if ( _datachain_begin == * __datachain_first && _datachain_end == * __datachain_last )
        { file_close();
          return;
        }
     if ( _datachain_begin == _datachain_end )
        { _datachain_next = _datachain_end->_next;
        }
     else
        { if ( insert_line_rq )
             { _datachain_next = _datachain_insert(_datasystem, _datachain_end);
             }
          else
             { _datachain_next = _datachain_end->_next;
             }
        }
     if ( _datachain_prev )
        { _datachain_prev->_next = _datachain_next;
        }
     else
        { * __datachain_first = _datachain_next;
        }
     if ( _datachain_next )
        { _datachain_next->_prev = _datachain_prev;
          * __datachain = _datachain_next;
        }
     else
        { * __datachain_last = _datachain_prev;
          cursor_up();
        }

        // -- free --
     Datachain * _datachain = _datachain_begin;
     Datachain * _datachain_tmp;
     Datachain * _datachain_end_next = _datachain_end->_next;
     do
        { _datachain_tmp = _datachain->_next;
          bookmark_remove(_datachain);
          datachain_free(_datachain);
          _datachain = _datachain_tmp;
        }
     while ( _datachain != _datachain_end_next );

        // -- ordinals --
     int delta = _datachain_end->ordinal - _datachain_begin->ordinal + 1;
     _datachain = _datachain_next;
     while ( _datachain )
        { _datachain->ordinal -= delta;
          _datachain = _datachain->_next;
        }
     datasystem_verify(_datasystem);

        // -- optimize screen --
     int visible_lines = visible_lines_up(* __datachain);
     if ( * _cursor_y > visible_lines + screen_top - 1 )
        { * _cursor_y = visible_lines + screen_top - 1;
        }
     * _status_register |= MODIFY_STATE;

     screen_redraw();

     return;
   }
void dataline_remove()
   { if ( ! (* _status_register & TYPE_MODE) )
        { return;
        }

     Datachain * _datachain = * __datachain;
     Datachain * _datachain_block_begin, * _datachain_block_end;

     int block_level = block_locate(* __datachain, & _datachain_block_begin, & _datachain_block_end);
     Datachain * _datachain_next = _datachain->_next;
     if ( block_level > 0 && _datachain_next && _datachain_next->invisible )
        { command_line_echo(_message[23]);
          int user_request = command_line_revolve(_confirm_str, 1, 2);
          if ( user_request == REPLY_YES )
             { block_remove(* __datachain, _datachain_block_end, 0);
             }
          return;
        }

        // -- single line --
     if ( _datachain->size )
        { _memset(_datachain->_line, ' ', _datachain->size);
          _datachain->size = 0;
          screen_line_redraw(_datachain, * _cursor_y);
          return;
        }
     block_remove(* __datachain, * __datachain, 0);
     return;
   }

void dataline_goto()
   { command_line_echo(_message[17]);
     command_line_read(_history_ptr[HISTORY_GOTO_LINE]);

     if ( ! (* _entry_line) )
        { return;
        }
     char * _user_line_point = _entry_line;
     while ( * _user_line_point )
        { if ( * _user_line_point < '0' || * _user_line_point > '9' )
             { return;
             }
          _user_line_point++;
        }
     int line_number = 0;
     line_number = atoi(_entry_line);
     # ifdef TEST_DATALINE_GOTO
        status_line_vprintf("%d", line_number);
     # endif
     Datachain * _datachain = * __datachain_first;
     while ( _datachain && _datachain->ordinal + 1 < line_number )
        { _datachain = _datachain->_next;
        }
     if ( ! _datachain )
        { status_line_vprintf(_message[36], (* __datachain_last)->ordinal + 1);
          return;
        }
     if ( _datachain->invisible )
        { status_line_vprintf(_message[37], line_number);
          return;
        }
     * __datachain = _datachain;
     * _dataline_offset = 0;
     * _cursor_y = (screen_lines - screen_top) / 2;
     if ( * _cursor_y > screen_top + visible_lines_up(* __datachain) - 1 )
        { * _cursor_y = screen_top + visible_lines_up(* __datachain) - 1;
        }
     * _cursor_x = screen_left;
     show_coords();
     screen_redraw();
     return;
   }
char * _dataline_advance(Datachain * _datachain, char * _dataline_point)
   { for ( int i = 0; i < _datachain->size; i++ )
         { if ( * _dataline_point != ' ' )
              { return _dataline_point;
              }
           _dataline_point++;
         }
     //return 0;
     return _dataline_point;
   }
bool dataline_is_empty(Datachain * _datachain)
   { if ( ! _datachain->size )
        { return 1;
        }
     char * _dataline_point = _dataline_advance(_datachain, _datachain->_line);
     if ( _dataline_point == _datachain->_line + _datachain->size )
        { return 1;
        }
     return 0;
   }
