# include "std.h"
# if LICENCE
   This file is part of cwrite.
   Copyright 2007, Desislav Georgiev.
 
   cwrite is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
 
   cwrite is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with cwrite.  If not, see <http://www.gnu.org/licenses/>.
# endif
# if HEADER
   # include <stdio.h>
   # include <stdlib.h>
   # include <fcntl.h>
   # include <unistd.h>

   # include "block.h"
   # include "command_line.h"
   # include "cwrite.h"
   # include "failure.h"
   # include "history.h"
   # include "io.h"
   # include "key.h"
   # include "message.h"
   # include "mycurses.h"
   # include "status_line.h"
   # include "string.h"
# endif
# if GLOBAL
   int history_entry_memsize = 80;
   History * _history_ptr[NUMBER_OF_HISTORIES];
# endif

void history_init()
   { for (int i = 0; i < NUMBER_OF_HISTORIES; i++)
        { _history_ptr[i] = (History *) calloc(1, sizeof(History));
          _history_ptr[i]->cardinal = 0;
             for (int j = 0; j < HISTORY_ENTRY_CARDINAL; j++) 
                { //_history_ptr[i]->_entry[j] = (char *) calloc(1, history_entry_memsize);
                }
        }
     return;
   }

int  history_set()
   { int history_size = 0;
     for (int i = 0; i < NUMBER_OF_HISTORIES; i++)
        { history_size += sizeof(_history_ptr[i]->cardinal);
          for (int j = 0; j < _history_ptr[i]->cardinal; j++)
             { _history_ptr[i]->entry_size[j] = string_len(_history_ptr[i]->_entry[j]);
               history_size += sizeof(_history_ptr[i]->entry_size[j]);
               history_size += _history_ptr[i]->entry_size[j];
             }
        }

     return history_size;
   }
void history_save()
   { int sid = 104;

     int fd = open(_user_history, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
     if ( fd == -1 )
        { fail(sid, _message[57]);
        }
     int history_size = history_set();
     io_write(fd, & history_size, sizeof(history_size));
     for (int i = 0; i < NUMBER_OF_HISTORIES; i++)
        { io_write(fd, & _history_ptr[i]->cardinal, sizeof(_history_ptr[i]->cardinal));
          for (int j = 0; j < _history_ptr[i]->cardinal; j++)
             { //_history_ptr[i]->entry_size[j] = string_len(_history_ptr[i]->_entry[j]);
               io_write(fd, & _history_ptr[i]->entry_size[j], sizeof(_history_ptr[i]->entry_size[j]));
               io_write(fd, _history_ptr[i]->_entry[j], _history_ptr[i]->entry_size[j]);
             }
        }
     io_close(fd);

     return;
   }
void history_load()
   { int sid = 114;

     int fd = open(_user_history, O_RDONLY);
     if ( fd == -1 )
        { return;
        }
     int history_size;
     int file_size = io_lseek(fd, 0, SEEK_END);
     if ( file_size < sizeof(history_size) )
        { io_close(fd);
          return;
        }
     io_lseek(fd, 0, SEEK_SET);
     io_read(fd, & history_size, sizeof(history_size));
     if ( history_size + sizeof(history_size) != file_size )
        { io_close(fd);
          return;
        }
     for (int i = 0; i < NUMBER_OF_HISTORIES; i++)
        { io_read(fd, & _history_ptr[i]->cardinal, sizeof(_history_ptr[i]->cardinal));
          for (int j = 0; j < _history_ptr[i]->cardinal; j++)
             { io_read(fd, & _history_ptr[i]->entry_size[j], sizeof(_history_ptr[i]->entry_size[j]));
               _history_ptr[i]->_entry[j] = (char *) calloc(1, _history_ptr[i]->entry_size[j] + 1);
               io_read(fd, _history_ptr[i]->_entry[j], _history_ptr[i]->entry_size[j]);
             }
        }
     io_close(fd);
   }

int  history_complete(History * _history)
   { if ( ! _history->cardinal )
        { return 1;
        }
     int ordinal = _history->cardinal;
     int key = KEY_UP;
     status_line_print(_message[42]);
     do
        { 
          switch(key)
             { case KEY_UP:
               case CTRL_P:
                  if ( ! ordinal )
                     {    // -- walk back if no match --
                       if ( * _entry_line && ! line_match(_history->_entry[ordinal], _entry_line_keep) )
                          { key = CTRL_N;
                            continue;
                          }
                       else
                          { break;
                          }
                     }
                  ordinal--;
                  if ( ! line_match(_history->_entry[ordinal], _entry_line_keep) )
                     { continue;
                     }
                  command_line_stop_print(_history->_entry[ordinal]);
                  break;
               case KEY_DOWN:
               case CTRL_N:
                  if ( ordinal == _history->cardinal - 1 )
                     { command_line_stop_print("%s", _entry_line);
                       return 1;
                     }
                  ordinal++;
                  if ( ! line_match(_history->_entry[ordinal], _entry_line_keep) )
                     { continue;
                     }
                  command_line_stop_print(_history->_entry[ordinal]);
                  break;
               default:
                  return key;
             }
          key = curses_getch();
        }
     while ( key && key != CTRL_M && key != CTRL_J);

     return key;
   }
void history_shift(History * _history)
   { char * _history_entry_tmp  = _history->_entry[0];
     int history_entry_size_tmp = _history->entry_size[0];

     for (int i = 0; i < HISTORY_ENTRY_CARDINAL-1; i++)
        { _history->_entry[i] = _history->_entry[i+1];
          _history->entry_size[i] = _history->entry_size[i+1];
        }
     _history->_entry[HISTORY_ENTRY_CARDINAL-1] = _history_entry_tmp;
     _history->entry_size[HISTORY_ENTRY_CARDINAL-1] = history_entry_size_tmp;
     _history->cardinal--;
     return;
   }
void history_update(History * _history)
   { int cardinal = _history->cardinal;
     int i;
     bool history_match = 0;
     int entry_line_size = string_len(_entry_line);
     for ( i = 0; i < cardinal; i++)
        { //if ( _history->entry_size[i] != entry_line_size )
          //   { continue;
          //   }
          if ( string_compare(_history->_entry[i], _entry_line) )
             { history_match = 1;
               break;   
             }
        }
     int ordinal = i;
     if ( history_match ) 
        { char * _history_entry_tmp  = _history->_entry[ordinal];
          int history_entry_size_tmp = _history->entry_size[ordinal];
          for (int i = ordinal; i < cardinal; i++)
             { _history->_entry[i]     = _history->_entry[i+1];
               _history->entry_size[i] = _history->entry_size[i+1];
             }
          _history->_entry[cardinal-1]     = _history_entry_tmp;
          _history->entry_size[cardinal-1] = history_entry_size_tmp;
        }
     else 
        { if ( _history->cardinal == HISTORY_ENTRY_CARDINAL )
             { history_shift(_history);
               cardinal = _history->cardinal;
             }
          if ( ! _history->_entry[cardinal] )
             { _history->_entry[cardinal] = (char *) calloc(1, entry_line_size + 1);
               _history->entry_size[cardinal] = entry_line_size;
             }
          if ( _history->entry_size[cardinal] < entry_line_size )
             { _history->_entry[cardinal] = (char *) realloc(_history->_entry[cardinal], entry_line_size + 1);
             }
          string_copy(_history->_entry[cardinal], _entry_line);
          _history->entry_size[cardinal] = entry_line_size;
          _history->cardinal++;
        }
     return;
   }
